@@ -137,7 +137,8 @@ struct Section {
     int32_t index;
     struct Reloc *head, **tail;
     uint32_t flags;             /* section flags */
-    char name[9];
+    char *name;
+    int32_t namepos;            /* Offset of name into the strings table */
     int32_t pos, relpos;
 };
 
@@ -215,7 +216,7 @@ static uint32_t strslen;
 static void coff_gen_init(void);
 static void coff_sect_write(struct Section *, const uint8_t *, uint32_t);
 static void coff_write(void);
-static void coff_section_header(char *, int32_t, int32_t, int32_t, int32_t, int, int32_t);
+static void coff_section_header(char *, int32_t, int32_t, int32_t, int32_t, int32_t, int, int32_t);
 static void coff_write_relocs(struct Section *);
 static void coff_write_symbols(void);
 
@@ -272,6 +273,7 @@ static void coff_cleanup(int debuginfo)
             sects[i]->head = sects[i]->head->next;
             nasm_free(r);
         }
+        nasm_free(sects[i]->name);
         nasm_free(sects[i]);
     }
     nasm_free(sects);
@@ -284,23 +286,31 @@ static void coff_cleanup(int debuginfo)
 static int coff_make_section(char *name, uint32_t flags)
 {
     struct Section *s;
+    size_t namelen;
 
-    s = nasm_malloc(sizeof(*s));
+    s = nasm_zalloc(sizeof(*s));
 
     if (flags != BSS_FLAGS)
         s->data = saa_init(1);
-    else
-        s->data = NULL;
-    s->head = NULL;
     s->tail = &s->head;
-    s->len = 0;
-    s->nrelocs = 0;
     if (!strcmp(name, ".text"))
         s->index = def_seg;
     else
         s->index = seg_alloc();
-    strncpy(s->name, name, 8);
-    s->name[8] = '\0';
+    s->namepos = -1;
+    namelen = strlen(name);
+    if (namelen > 8) {
+        if (win32 || win64) {
+            s->namepos = strslen + 4;
+            saa_wbytes(strs, name, namelen + 1);
+            strslen += namelen + 1;
+        } else {
+            namelen = 8;
+        }
+    }
+    s->name = nasm_malloc(namelen + 1);
+    strncpy(s->name, name, namelen);
+    s->name[namelen] = '\0';
     s->flags = flags;
 
     if (nsects >= sectlen) {
@@ -342,9 +352,11 @@ static int32_t coff_section_names(char *name, int pass, int *bits)
     if (*p)
         *p++ = '\0';
     if (strlen(name) > 8) {
-        nasm_error(ERR_WARNING, "COFF section names limited to 8 characters:"
-              " truncating");
-        name[8] = '\0';
+        if (!win32 && !win64) {
+            nasm_error(ERR_WARNING,
+                       "COFF section names limited to 8 characters:  truncating");
+            name[8] = '\0';
+        }
     }
     flags = 0;
 
@@ -430,9 +442,42 @@ static int32_t coff_section_names(char *name, int pass, int *bits)
         sects[i]->flags &= align_and;
         sects[i]->flags |= align_or;
     } else if (pass == 1) {
-        if (flags)
-            nasm_error(ERR_WARNING, "section attributes ignored on"
-                  " redeclaration of section `%s'", name);
+        /* Check if any flags are specified */
+        if (flags) {
+            unsigned int align_flags = flags & IMAGE_SCN_ALIGN_MASK;
+
+            /* Warn if non-alignment flags differ */
+            if ((flags ^ sects[i]->flags) & ~IMAGE_SCN_ALIGN_MASK) {
+                nasm_error(ERR_WARNING, "section attributes ignored on"
+                    " redeclaration of section `%s'", name);
+            }
+            /* Check if alignment might be needed */
+            if (align_flags > IMAGE_SCN_ALIGN_1BYTES) {
+                unsigned int sect_align_flags = sects[i]->flags & IMAGE_SCN_ALIGN_MASK;
+
+                /* Compute the actual alignment */
+                unsigned int align = 1u << ((align_flags - IMAGE_SCN_ALIGN_1BYTES) >> 20);
+
+                /* Update section header as needed */
+                if (align_flags > sect_align_flags) {
+                    sects[i]->flags = (sects[i]->flags & ~IMAGE_SCN_ALIGN_MASK) | align_flags;
+                }
+                /* Check if not already aligned */
+                if (sects[i]->len % align) {
+                    unsigned int padding = (align - sects[i]->len) % align;
+                    /* We need to write at most 8095 bytes */
+                    char buffer[8095];
+                    if (sects[i]->flags & IMAGE_SCN_CNT_CODE) {
+                        /* Fill with INT 3 instructions */
+                        memset(buffer, 0xCC, padding);
+                    } else {
+                        memset(buffer, 0x00, padding);
+                    }
+                    saa_wbytes(sects[i]->data, buffer, padding);
+                    sects[i]->len += padding;
+                }
+            }
+        }
     }
 
     return sects[i]->index;
@@ -919,7 +964,7 @@ static void coff_write(void)
      */
     vsize = 0L;
     for (i = 0; i < nsects; i++) {
-        coff_section_header(sects[i]->name, vsize, sects[i]->len,
+        coff_section_header(sects[i]->name, sects[i]->namepos, vsize, sects[i]->len,
                             sects[i]->pos, sects[i]->relpos,
                             sects[i]->nrelocs, sects[i]->flags);
         vsize += sects[i]->len;
@@ -942,7 +987,7 @@ static void coff_write(void)
     saa_fpwrite(strs, ofile);
 }
 
-static void coff_section_header(char *name, int32_t vsize,
+static void coff_section_header(char *name, int32_t namepos, int32_t vsize,
                                 int32_t datalen, int32_t datapos,
                                 int32_t relpos, int nrelocs, int32_t flags)
 {
@@ -950,9 +995,32 @@ static void coff_section_header(char *name, int32_t vsize,
 
     (void)vsize;
 
-    memset(padname, 0, 8);
-    strncpy(padname, name, 8);
-    fwrite(padname, 8, 1, ofile);
+    if (namepos == -1) {
+        strncpy(padname, name, 8);
+        fwrite(padname, 8, 1, ofile);
+    } else {
+        /*
+         * If name is longer than 8 bytes, write '/' followed
+         * by offset into the strings table represented as
+         * decimal number.
+         */
+        namepos = namepos % 100000000;
+        padname[0] = '/';
+        padname[1] = '0' + (namepos / 1000000);
+        namepos = namepos % 1000000;
+        padname[2] = '0' + (namepos / 100000);
+        namepos = namepos % 100000;
+        padname[3] = '0' + (namepos / 10000);
+        namepos = namepos % 10000;
+        padname[4] = '0' + (namepos / 1000);
+        namepos = namepos % 1000;
+        padname[5] = '0' + (namepos / 100);
+        namepos = namepos % 100;
+        padname[6] = '0' + (namepos / 10);
+        namepos = namepos % 10;
+        padname[7] = '0' + (namepos);
+        fwrite(padname, 8, 1, ofile);
+    }
 
     fwriteint32_t(0,            ofile); /* Virtual size field - set to 0 or vsize */
     fwriteint32_t(0L,           ofile); /* RVA/offset - we ignore */
@@ -1003,7 +1071,6 @@ static void coff_symbol(char *name, int32_t strpos, int32_t value,
     char padname[8];
 
     if (name) {
-        memset(padname, 0, 8);
         strncpy(padname, name, 8);
         fwrite(padname, 8, 1, ofile);
     } else {
@@ -1028,7 +1095,6 @@ static void coff_write_symbols(void)
      * The `.file' record, and the file name auxiliary record.
      */
     coff_symbol(".file", 0L, 0L, -2, 0, 0x67, 1);
-    memset(filename, 0, 18);
     strncpy(filename, coff_infile, 18);
     fwrite(filename, 18, 1, ofile);
 
